#!perl -w
use Test::More tests => 24;
use strict;
use File::Temp qw( tempdir );
use File::Basename qw(basename);
use Time::HiRes qw (time);

use DE_EPAGES::Core::API::File   qw( DeleteDirectory CopyDirectory GetFileNames
                                     CreateDirectory CompareDirectories WriteFile
                                     RunOnDirectory GetFileContent );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::Error  qw( GetError ExistsError );

sub Main {
  SKIP: {
      skip('CopyDirectory(..., {by => \'exe\'}): not implemented on Windows.', 23) if $^O eq 'MSWin32';

      my $TempDir = tempdir( CLEANUP => 1 );
      foreach (qw(11 111 112 121 131 21 211 213 221 222)) {
          WriteFile(sprintf("$TempDir/%s.txt", join('/',split(//,$_))), \$_, {create_directory=>1});
      }
      my $Content = '1.1';
      WriteFile("$TempDir/1/.1.txt", \$Content);
      CreateDirectory("$TempDir/3");
      sleep 1; # source and target shall not have the same timestamp

      # test 1: copy with perl
      CopyDirectory("$TempDir/1", "$TempDir/11", undef, {by => 'perl', Print => 1});
      ok(CompareDirectories("$TempDir/1", "$TempDir/11"), "CopyDirectory($TempDir/1, $TempDir/11, undef, {by => 'perl'})");

      # test 2: copy with perl
      CopyDirectory("$TempDir/2", "$TempDir/E", undef, {by => 'perl', Print => 1});
      ok(CompareDirectories("$TempDir/2", "$TempDir/E"), "CopyDirectory($TempDir/2, $TempDir/E, undef, {by => 'perl'})");

      # test 3: copy with perl
      CopyDirectory("$TempDir/3", "$TempDir/F", undef, {by => 'perl', Print => 1});
      ok(CompareDirectories("$TempDir/3", "$TempDir/F"), "CopyDirectory($TempDir/3, $TempDir/F, undef, {by => 'perl'})");

      # test 4/5: copy with /bin/cp (stat => 'ignore')
      CopyDirectory("$TempDir/1", "$TempDir/12", undef, {by => 'exe', stat => 'ignore', Print => 1});
      my $hInodes = {};
      _GetInodes("$TempDir/1", $hInodes);
      _GetInodes("$TempDir/12", $hInodes);
      ok(CompareDirectories("$TempDir/1", "$TempDir/12"), "CopyDirectory($TempDir/1, $TempDir/12)");
      ok(_CompareStat($hInodes, 1, '12'), "CopyDirectory($TempDir/1, $TempDir/12, undef, {by => 'exe', stat => 'ignore'})");

      # test 6: copy with /bin/cp (stat => 'preserve')
      CopyDirectory("$TempDir/1", "$TempDir/13", undef, {by => 'exe', stat => 'preserve', Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/1", $hInodes);
      _GetInodes("$TempDir/13", $hInodes);
      ok(_CompareStat($hInodes, 1, '13', 'preserve'), "CopyDirectory($TempDir/1, $TempDir/13, undef, {by => 'exe', stat => 'preserve'})");

      # test 7: copy with default (perl): stat => 'epages'
      # stat is preserved because target is located outside epages
      CopyDirectory("$TempDir/1", "$TempDir/14", undef, {Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/1", $hInodes);
      _GetInodes("$TempDir/14", $hInodes);
      ok(_CompareStat($hInodes, 1, '14', 'preserve'), "CopyDirectory($TempDir/1, $TempDir/14, undef, { stat => 'epages'})");

    SKIP: {
        skip(sprintf('User \'%s\' can\'t set file permissions, only \'root\' can do this.', (getpwuid $<)[0]), 3) if $< != 0;

        # test 8: copy with default (perl): stat => 'epages'
        # stat is epages because target is located inside and source outside epages
        DeleteDirectory("$ENV{EPAGES}/bin/15", {recursive => 1});
        CopyDirectory("$TempDir/1", "$ENV{EPAGES}/bin/15", undef, {Print => 1});
        sleep 1; # source and target shall not have the same timestamp
        $hInodes = {};
        _GetInodes("$TempDir/1", $hInodes);
        _GetInodes("$ENV{EPAGES}/bin/15", $hInodes);
        ok(_CompareStat($hInodes, 1, '15', 'epages'), "CopyDirectory($TempDir/1, $ENV{EPAGES}/bin/15, undef, { stat => 'epages'})");

        # test 9: copy with default (perl): stat => 'epages'
        # stat is epages because source and target are located in different parts of epages
        DeleteDirectory("$ENV{EPAGES}/Setup/16", {recursive => 1});
        CopyDirectory("$ENV{EPAGES}/bin/15", "$ENV{EPAGES}/Setup/16", undef, {Print => 1});
        sleep 1; # source and target shall not have the same timestamp
        $hInodes = {};
        _GetInodes("$ENV{EPAGES}/bin/15", $hInodes);
        _GetInodes("$ENV{EPAGES}/Setup/16", $hInodes);
        ok(_CompareStat($hInodes, '15', '16', 'epages'), "CopyDirectory($ENV{EPAGES}/bin/15, $ENV{EPAGES}/Setup/16, undef, { stat => 'epages'})");

        # test 10: copy with default (perl): stat => 'epages'
        # stat is preserve because source and target are located in same parts of epages
        DeleteDirectory("$ENV{EPAGES}/Setup/17", {recursive => 1});
        CopyDirectory("$ENV{EPAGES}/Setup/16", "$ENV{EPAGES}/Setup/17", undef, {Print => 1});
        $hInodes = {};
        _GetInodes("$ENV{EPAGES}/Setup/16", $hInodes);
        _GetInodes("$ENV{EPAGES}/Setup/17", $hInodes);
        ok(_CompareStat($hInodes, '16', '17', 'preserve'), "CopyDirectory($ENV{EPAGES}/Setup/16, $ENV{EPAGES}/Setup/17, undef, { stat => 'epages'})");

        DeleteDirectory("$ENV{EPAGES}/bin/15", {recursive => 1});
        DeleteDirectory("$ENV{EPAGES}/Setup/16", {recursive => 1});
        DeleteDirectory("$ENV{EPAGES}/Setup/17", {recursive => 1});
      }

      # test 11: copy with /bin/cp
      CopyDirectory("$TempDir/2", "$TempDir/B", undef, {by => 'exe', Print => 1});
      ok(CompareDirectories("$TempDir/2", "$TempDir/B"), "CopyDirectory($TempDir/2, $TempDir/B, {by => 'exe'})");

      # test 12: copy with /bin/cp
      CopyDirectory("$TempDir/3", "$TempDir/C", undef, {by => 'exe', Print => 1});
      ok(CompareDirectories("$TempDir/3", "$TempDir/C"), "CopyDirectory($TempDir/3, $TempDir/C, {by => 'exe'})");

      # test 13: first copy, then copy
      CopyDirectory("$TempDir/1", "$TempDir/a", undef, {Print => 1});
      CopyDirectory("$TempDir/2", "$TempDir/a", undef, {Print => 1});
      ok(CompareDirectories("$TempDir/1", "$TempDir/a") && CompareDirectories("$TempDir/2", "$TempDir/a"),
         'CopyDirectory: first copy, then copy');

      # test 14: first copy, then link
      CopyDirectory("$TempDir/1", "$TempDir/b", undef, {Print => 1});
      CopyDirectory("$TempDir/2", "$TempDir/b", undef, {link => 1, by => 'exe', Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/1", $hInodes);
      _GetInodes("$TempDir/2", $hInodes);
      _GetInodes("$TempDir/b", $hInodes);
      ok(_CompareInodes($hInodes, undef, 'link'), 'CopyDirectory: first copy, then link');

      # test 15: first link, then copy
      CopyDirectory("$TempDir/1", "$TempDir/c", undef, {link => 1, by => 'exe', Print => 1});
      CopyDirectory("$TempDir/2", "$TempDir/c", undef, {Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/1", $hInodes);
      _GetInodes("$TempDir/2", $hInodes);
      _GetInodes("$TempDir/c", $hInodes);
      ok(_CompareInodes($hInodes, 'link'), 'CopyDirectory: first link, then copy');

      # test 16: first link, then link
      # we need new $Copy1 because it was changed in test 9
      CopyDirectory("$TempDir/11", "$TempDir/d", undef, {link => 1, by => 'exe', Print => 1});
      CopyDirectory("$TempDir/2", "$TempDir/d", undef, {link => 1, by => 'exe', Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/11", $hInodes);
      _GetInodes("$TempDir/2", $hInodes);
      _GetInodes("$TempDir/d", $hInodes);
      ok(_CompareInodes($hInodes, 'link', 'link'), 'CopyDirectory: first link, then link');

      # test 17: after link, can't copy same again
      my $Error;
      eval { CopyDirectory("$TempDir/2", "$TempDir/d", undef, {by => 'exe', Print => 1}); };
      $Error = GetError() if ExistsError();
      ok($Error->message =~ /the same file/,
         'CopyDirectory: after link, copy same again -> correct error message: the same file');

      # test 18: but it works with unlink (target has new inodes)
      CopyDirectory("$TempDir/2", "$TempDir/d", undef, {unlink => 1, Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/11", $hInodes);
      _GetInodes("$TempDir/2", $hInodes);
      _GetInodes("$TempDir/d", $hInodes);
      ok(_CompareInodes($hInodes, 'link', 'unlink'),
         'CopyDirectory: after link, copy same again -> works with unlink');

      # test 19: but it works with unlink/link (target has same inodes)
      CopyDirectory("$TempDir/2", "$TempDir/d", undef, {unlink => 1, link => 1, by => 'exe', Print => 1});
      $hInodes = {};
      _GetInodes("$TempDir/11", $hInodes);
      _GetInodes("$TempDir/2", $hInodes);
      _GetInodes("$TempDir/d", $hInodes);
      ok(_CompareInodes($hInodes, 'link', 'link'),
         'CopyDirectory: after link, link same again -> works with unlink');

      # test 20: cannot handle symlinks with /bin/cp
      DeleteDirectory("$TempDir/d/1", {recursive => 1});
      symlink("$TempDir/d/2", "$TempDir/d/1");
      eval { CopyDirectory("$TempDir/2", "$TempDir/d", undef,
                           {unlink => 1, by => 'exe', Print => 1}); };
      $Error = GetError() if ExistsError();
      ok($Error->message =~ /cannot overwrite non-directory/,
         'CopyDirectory: copy directory symlink -> correct error message: /bin/cp: cannot overwrite non-directory');

      # test 21: but it works with perl
      eval { CopyDirectory("$TempDir/2", "$TempDir/d", undef,
                           {unlink => 1, by => 'perl', Print => 1}); };
      ok(ExistsError() ? 0 : 1,
         'CopyDirectory: copy directory symlink -> works with File::Copy');

      SKIP: {
          skip( 'Forced skip of timing tests.', 3 ) if $ENV{'SKIP_TIMING'};

          DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp", {recursive => 1});
          my $Duration = time;
          CopyDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation", "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp",
                    undef, {by => 'perl', Print => 1});
          $Duration =  time - $Duration;

          # test 22: compare perl link and perl copy
          DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp1", {recursive => 1});
          my $Duration1 = time;
          CopyDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation", "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp1",
                    undef, {link => 1, by => 'perl', Print => 1});
          $Duration1 =  time - $Duration1;
          ok($Duration1 < $Duration, sprintf('CopyDirectory: perl link (%.3f sec) is faster than perl copy (%.3f sec)', $Duration1, $Duration));

          # test 23: compare /bin/cp link and perl link
          DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp3", {recursive => 1});
          my $Duration3 = time;
          CopyDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation", "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp3",
                    undef, {link => 1, by => 'exe', Print => 1});
          $Duration3 =  time - $Duration3;
          ok($Duration3 < $Duration1, sprintf('CopyDirectory: /bin/cp link (%.3f sec) is faster than perl link (%.3f sec)', $Duration3, $Duration1));
      }

      DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp", {recursive => 1});
      DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp1", {recursive => 1});
      DeleteDirectory("$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation.tmp3", {recursive => 1});
    }
}

RunScript(Sub=>\&Main);

sub _GetInodes ($$) {
    my ($Path, $hInodes) = @_;
    my $Dir = basename($Path);
    $Dir = 'T' unless $Dir =~ /^[0-9]+$/;
    foreach my $File (@{RunOnDirectory($Path, sub { return GetFileNames(undef, {'recursive' => 1}); })}) {
        $hInodes->{$File}{$Dir}{inode} = (stat "$Path/$File")[1];
        $hInodes->{$File}{$Dir}{gid} = (stat "$Path/$File")[5];
        $hInodes->{$File}{$Dir}{mtime} = (stat "$Path/$File")[9];
        $hInodes->{$File}{$Dir}{ctime} = (stat "$Path/$File")[10];
        my $sContent = GetFileContent("$Path/$File");
        $hInodes->{$File}{$Dir}{content} = $$sContent;
    }
}

sub _CompareStat ($;$$) {
    my ($hInodes, $Source, $Target, $Stat) = @_;
    my $Gid = (getgrnam 'epages')[2];

    foreach my $File (keys %$hInodes) {
        if ( $Stat eq 'preserve' ) {
            # error if source and target have different mtime
            return 0 if $hInodes->{$File}{$Source}{mtime} != $hInodes->{$File}{$Target}{mtime};
        } elsif ( $Stat eq 'epages' ) {
            # error if source and target have same ctime
            return 0 if $hInodes->{$File}{$Source}{ctime} == $hInodes->{$File}{$Target}{ctime};
            return 0 if $Gid != $hInodes->{$File}{$Target}{gid};
        } else {
            # error if source and target have same mtime
            return 0 if $hInodes->{$File}{$Source}{mtime} == $hInodes->{$File}{$Target}{mtime};
        }
    }
    return 1;
}

sub _CompareInodes ($;$$) {
    my ($hInodes, $Copy1, $Copy2) = @_;

    my $cp1 = 1;
    foreach my $File (keys %$hInodes) {
        foreach (keys %{$hInodes->{$File}}) {
            if (/^1/) { $cp1 = $_; last; }
        }
    }

    foreach my $File (keys %$hInodes) {
        if (not $Copy1 and $Copy2 eq 'link') {
            # first copy, then link
            if ( defined $hInodes->{$File}{2} ) {
                # error if target has different inode than $Copy2 (link)
                return 0 if $hInodes->{$File}{T}{inode} != $hInodes->{$File}{2}{inode};
            } else {
                # error if target has same inode than $Copy1 (copy)
                return 0 if $hInodes->{$File}{T}{inode} == $hInodes->{$File}{$cp1}{inode};
                # error if target content and $Copy1 content differs
                return 0 if $hInodes->{$File}{T}{content} ne $hInodes->{$File}{$cp1}{content};
            }
        }
        elsif ($Copy1 eq 'link' and not $Copy2) {
            if ( defined $hInodes->{$File}{2} ) {
                # the second run overwrites both $Copy1 and target because they are hard-linked
                # error if target has same inode than $Copy2 (copy)
                return 0 if $hInodes->{$File}{T}{inode} == $hInodes->{$File}{2}{inode};
                # error if target content and $Copy2 content differs
                return 0 if $hInodes->{$File}{T}{content} ne $hInodes->{$File}{2}{content};
            }
            if ( defined $hInodes->{$File}{$cp1} ) {
                # error if target has different inode than $Copy1 (link)
                return 0 if $hInodes->{$File}{T}{inode} != $hInodes->{$File}{$cp1}{inode};
            }
        }
        elsif ($Copy1 eq 'link' and $Copy2 eq 'unlink') {
            if ( defined $hInodes->{$File}{2} ) {
                # the second run overwrites both $Copy1 and target because they are hard-linked
                # error if target has same inode than $Copy2 (copy)
                return 0 if $hInodes->{$File}{T}{inode} == $hInodes->{$File}{2}{inode};
                # error if target content and $Copy2 content differs
                return 0 if $hInodes->{$File}{T}{content} ne $hInodes->{$File}{2}{content};
            } elsif ( defined $hInodes->{$File}{$cp1} ) {
                # error if target has different inode than $Copy1 (link)
                return 0 if $hInodes->{$File}{T}{inode} != $hInodes->{$File}{$cp1}{inode};
            }
        }
        elsif ($Copy1 eq 'link' and $Copy2 eq 'link' ) {
            if ( defined $hInodes->{$File}{2} ) {
                # error if target has different inode than $Copy2 (link)
                return 0 if $hInodes->{$File}{T}{inode} != $hInodes->{$File}{2}{inode};
            } else {
                # error if target has different inode than $Copy1 (link)
                return 0 if $hInodes->{$File}{T}{inode} != $hInodes->{$File}{$cp1}{inode};
            }
        }
    }
    return 1;
}

1;
