use Test::More;
use Test::Deep;

use strict;
use SOAP::Lite;

use DE_EPAGES::Core::API::File qw ( OpenFile DeleteFile );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw (
    GetTestSite
    RunOnStore
    Transaction
);
use DE_EPAGES::Object::API::Factory qw (
    LoadObjectByPath
);
use DE_EPAGES::Shop::API::Shop qw ( GetNextNo );

use Readonly;
Readonly::Scalar my $BASE_FILENAME => 'GetNextNo';
Readonly::Scalar my $NUM_TASKS     => 8;
Readonly::Scalar my $ROOTPATH      => '/Shops';

# Cannot use /Test/TestShop, because every attempt to access the Store database
# prior to invoking the subroutine GetNextCustomerNo* results in MySQL errors:
# Mysql complains that AutoCommit is not set, although it is set.
local $ENV{'EPAGES_TESTSITE'} = 'Store';

#========================================================================================
# description  Tests simultaneous access to the subroutine $FunctionRef by creating
#               several child processes, each calling $FunctionRef in parallel. That
#               way, race conditions that might be raised during the execution of
#               $FunctionRef can be detected. We assume that the subroutine
#               'IncrementIntegerAttribute' (cf. Object/DAL/SQL/DefaultAttribute.pm)
#               is already thread-safe.
#========================================================================================
sub TestGetNextNo {
    my ($Identifier, $FunctionRef) = @_;

    my @ForkedTasks;
    my $MaxNumbers = 1; ## no critic qw(ProhibitMagicNumbers)

    for (my $Count = 1; $Count <= $NUM_TASKS; $Count++) {
        my $PID = fork();
        if ($PID) {  # parent
            push @ForkedTasks, $PID;
        } elsif ($PID == 0) {  # child
            my $NextNumbers = $FunctionRef->($Count, $MaxNumbers);

            ok( scalar @{ $NextNumbers } > 0,
                "new numbers were generated by child $Count ($Identifier)" );
            is( scalar @{ $NextNumbers }, $MaxNumbers,
                "$MaxNumbers new numbers were generated by child $Count ($Identifier)" );

            OpenFile('>>', "$BASE_FILENAME$Count", sub {
                my ($FH) = @_;
                print $FH "$_\n" foreach @{ $NextNumbers };
            });

            exit 0;
        } else {
            die "Couldn't fork: $!\n";
        }
    }

    foreach my $Task ( @ForkedTasks ) {
        my $KID;
        do {
            $KID = waitpid($Task, 0);
        } while $KID > 0;
    }

    my @AllNumbers;
    for (my $Count = 1; $Count <= $NUM_TASKS; $Count++) {
        OpenFile('<', "$BASE_FILENAME$Count", sub {
            my ($FH) = @_;
            while (<$FH>) {
                chomp;
                push @AllNumbers, $_;
            }
        });
    }

    my %Set = map { $_ => 1 } @AllNumbers;
    cmp_deeply( \@AllNumbers, bag(keys %Set),
        "bag of generated numbers from all processes is unique ($Identifier)" );

    return;
}

#========================================================================================
# description  Invokes the subroutine 'GetNextNo' in order to retrieve the next
#               customer number.
#========================================================================================
sub GetNextCustomerNoTask {
    my ($Count, $MaxNumbers) = @_;

    my @Numbers;
    RunOnStore(
        Store => GetTestSite(),
        Sub   => sub {
            my $Shop = LoadObjectByPath($ROOTPATH)->child('DemoShop');
            push @Numbers, GetNextNo($Shop, 'Customer');
            sleep(rand(3));  # let the other child processes interfere
        }
    );

    return \@Numbers;
}

#========================================================================================
# description  Invokes the subroutine 'GetNextNo' in order to retrieve the next
#               customer number; nested by a Transaction.
#========================================================================================
sub GetNextCustomerNoTaskByTransaction {
    my ($Count, $MaxNumbers) = @_;

    my @Numbers;
    RunOnStore(
        Store => GetTestSite(),
        Sub   => sub {
            Transaction(sub {
                my $Shop = LoadObjectByPath($ROOTPATH)->child('DemoShop');
                push @Numbers, GetNextNo($Shop, 'Customer');
                sleep(rand(3));  # let the other child processes interfere
            });
        }
    );

    return \@Numbers;
}

sub Main {
    ok( Init(), 'intialization done' );

    SKIP: {
        # There seems to be an issue while using 'fork' in connection with Windows
        skip( 'Tests produce errors in Windows.', 4 ) if $^O eq 'MSWin32';

        TestGetNextNo('NextCustomerNo', \&GetNextCustomerNoTask);
        TestGetNextNo('NextCustomerNoWithTransaction', \&GetNextCustomerNoTaskByTransaction);
    }

    ok( Cleanup(), 'cleanup done' );
    done_testing();

    return;
}

sub Init {
    Cleanup();

    return 1;
}

#========================================================================================
# description  Issue: Cannot reset NextCustomerNo to original value.
#========================================================================================
sub Cleanup {
    for (my $Count = 1; $Count <= $NUM_TASKS; $Count++) {
        DeleteFile("$BASE_FILENAME$Count");
    }

    return 1;
}

RunScript(
    Message => 'long',
    Sub     => \&Main,
);

1;