use Test::More tests => 8;

use strict;

use DE_EPAGES::Core::API::Script qw (RunScript);
use DE_EPAGES::Database::API::Connection qw ( GetTestSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject LoadObjectByPath ExistsObjectByPath );
use DE_EPAGES::Customer::API::Shop qw ( GetCustomerFolder GetNextCustomerNo );

my $ROOTPATH = '/Test';

#========================================================================================
# description  Tests the correctness of assigning always the next available alias
#               (=NextCustomerNo) when creating a new customer. The database stores the
#               next available (i.e. incremented) customer number, which can
#               be retrieved via $Shop->get('NextCustomerNo'). GetNextCustomerNo
#               increments this value by 1 as long as the newly incremented number is
#               not used.
#========================================================================================
sub TestNewCustomer {
    my $Shop = LoadObjectByPath($ROOTPATH)->child('TestShop');
    ok( $Shop->id, 'created shop ' . $Shop->alias );

    my $CustomerFolder = GetCustomerFolder($Shop);
    my $CustomerClass = $Shop->get('CustomerClass');

    # add two customers having the aliases 1001 and 1002
    $CustomerClass->insertObject({
        Alias  => $_,
        Parent => $CustomerFolder,
    }) foreach ( qw( 1001 1002 ) );

    # since the default value for NextCustomerNo equals 1000, and 1001 and 1002
    # are already assigned, GetNextCustomerNo should return 1000
    my $DatabaseAlias = $Shop->get('NextCustomerNo');
    ## no critic qw(ProhibitMagicNumbers)
    is( $DatabaseAlias, 1000, 'value in database equals default value of \'1000\'' );

    # next alias: 1000
    my $Customer = $CustomerClass->insertObject({
        Alias  => GetNextCustomerNo($Shop),
        Parent => $CustomerFolder,
    });
    is( $Customer->alias, 1000, 'new customer received alias \'1000\'' );
    $DatabaseAlias = $Shop->get('NextCustomerNo');
    is( $DatabaseAlias, 1001, 'value in database equals \'1001\'' );

    # next alias: 1003, since 1000, 1001, and 1002 are already used
    $Customer = $CustomerClass->insertObject({
        Alias  => GetNextCustomerNo($Shop),
        Parent => $CustomerFolder,
    });
    is( $Customer->alias, 1003, 'new customer received alias \'1003\'' );
    $DatabaseAlias = $Shop->get('NextCustomerNo');
    is( $DatabaseAlias, 1004, 'value in database equals \'1004\'' );
    ## use critic
}

sub Cleanup {
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);

    return 1;
}

sub Init {
    Cleanup();

    my $Folder = LoadRootObject()->insertChild('Object', {
        Alias => 'Test',
    });
    $Folder->insertChild('Shop', {
        Alias => 'TestShop',
    });

    return 1;
}

RunScript(
    Message => 'long',
    Sub     => sub {
        RunOnStore(
            Store => GetTestSite(),
            Sub   => sub {
                ok( Init(), 'intialization done' );

                TestNewCustomer();

                ok( Cleanup(), 'cleanup done' );
            },
        );
    }
);
